CREATE TABLE `multi1` (
  `name` varchar(30) DEFAULT NULL
) ENGINE=InnoDB DEFAULT CHARSET=latin1;

CREATE TABLE `multi2` (
  `id` int(10) unsigned DEFAULT NULL
) ENGINE=InnoDB DEFAULT CHARSET=latin1; CREATE TABLE `multi3` (
  `whatever` char(3) DEFAULT NULL
) ENGINE=InnoDB DEFAULT CHARSET=latin1
;