CREATE TABLE one (
	id int unsigned NOT NULL,
	PRIMARY KEY (id)
);

CREATE TABLE glorp.two (
	id int unsigned NOT NULL,
	PRIMARY KEY (id)
);

use bar;
CREATE TABLE three (
	id int unsigned NOT NULL,
	PRIMARY KEY (id)
);
